/***************************************************************************
 *   Copyright (C) 2005 by Robot Group Leipzig                             *
 *    martius@informatik.uni-leipzig.de                                    *
 *    fhesse@informatik.uni-leipzig.de                                     *
 *    der@informatik.uni-leipzig.de                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************
 *                                                                         *
 *  DESCRIPTION                                                            *
 *                                                                         *
 *   $Log: feedforwardnn.h,v $
 *   Revision 1.5  2006/07/20 17:14:34  martius
 *   removed std namespace from matrix.h
 *   storable interface
 *   abstract model and invertablemodel as superclasses for networks
 *
 *   Revision 1.4  2006/07/18 14:49:48  martius
 *   invertable networks provide the linear response function (jacobian)
 *
 *                                                                 *
 ***************************************************************************/
#ifndef __FEEDFORWARDNN_H
#define __FEEDFORWARDNN_H

#include "abstractmodel.h"
#include <math.h>

typedef double (*ActivationFunction) (double);

/// abstract class (interface) for feed forward rate based neural networks
class FeedForwardNN : public AbstractModel {
 public: 
  FeedForwardNN(){};
  virtual ~FeedForwardNN(){};

  /// damps the weights and the biases by multiplying (1-damping)
  virtual void damp(double damping) =0 ;

  static double linear(double x)   { return x;}
  static double dlinear(double )   { return 1;}
  static double tanh(double x)     { return ::tanh(x); } 
  static double dtanh(double x)    { double k = ::tanh(x); return 1.01-k*k; } 
  static double sigmoid(double x)  { return 1/(1+exp(-x)); } 
  static double dsigmoid(double x) { double k = sigmoid(x); return k*(1-k); } 

};


#endif
