/***************************************************************************
 *   Copyright (C) 2005 by Robot Group Leipzig                             *
 *    martius@informatik.uni-leipzig.de                                    *
 *    fhesse@informatik.uni-leipzig.de                                     *
 *    der@informatik.uni-leipzig.de                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 *   $Log: positionsensor.h,v $
 *   Revision 1.1  2005/11/22 10:24:04  martius
 *   abstract class for position sensor
 *
 *                                                                         *
 ***************************************************************************/
#ifndef __POSITIONSENSOR_H
#define __POSITIONSENSOR_H

#include <list>
using namespace std;
#include <ode/common.h>
#include "position.h"

/** Abstract class for Position-sensors. 
    The sensor value is usually some kind of distance to some fixed object in space
 */
class PositionSensor {
public:  
  PositionSensor() {}
  virtual ~PositionSensor() {}
  
  /** initialises sensor with body of robot
      @return number of sensor values returned by get
   */
  virtual int init(dBodyID body) = 0;  
  
  /** performs sense action
   */
  virtual bool sense(const GlobalData& globaldata) = 0;

  /** returns a list of sensor values (usually in the range [0,1] )
   */
  virtual list<double> get() = 0;

};

#endif
