package unileipzig.isys.neuronviz.app;

import java.awt.BorderLayout;
import java.awt.Dimension;
import javax.swing.JFrame;

/**
 * Class NeuronVizFrame
 * 
 */
public class NeuronVizFrame extends JFrame
{
	private String frameTitle = "Neuron Visualizer";
	private Dimension frameDimension = new Dimension(600, 400);

	public NeuronVizFrame()
	{
		setTitle(frameTitle);
		setSize(frameDimension);
		setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		
		getContentPane().setLayout(new BorderLayout());
	}
}
